// src/define-model/transform.ts
import { compileScript } from "@vue/compiler-sfc";
import MagicString from "magic-string";
import { extractIdentifiers, walkAST } from "ast-walker-scope";

// ../common/src/ast.ts
import { parse, walkIdentifiers } from "@vue/compiler-sfc";
var parseSFC = (code, id) => {
  const { descriptor } = parse(code, {
    filename: id
  });
  return descriptor;
};
function isCallOf(node, test) {
  return !!(node && node.type === "CallExpression" && node.callee.type === "Identifier" && (typeof test === "string" ? node.callee.name === test : test(node.callee.name)));
}

// ../common/src/constants.ts
var DEFINE_PROPS = "defineProps";
var DEFINE_EMITS = "defineEmits";
var DEFINE_MODEL = "defineModel";

// src/define-model/transform.ts
var transformDefineModel = (code, filename, version) => {
  let hasDefineProps = false;
  let hasDefineEmits = false;
  let hasDefineModel = false;
  let propsTypeDecl;
  let propsDestructureDecl;
  let emitsTypeDecl;
  let emitsIdentifier;
  let modelTypeDecl;
  let modelIdentifier;
  let modelDeclKind;
  let modelDestructureDecl;
  const modelIdentifiers = /* @__PURE__ */ new Set();
  function processDefinePropsOrEmits(node, declId) {
    var _a, _b;
    let type;
    if (isCallOf(node, DEFINE_PROPS)) {
      type = "props";
    } else if (isCallOf(node, DEFINE_EMITS)) {
      type = "emits";
    } else {
      return false;
    }
    const fnName = type === "props" ? DEFINE_PROPS : DEFINE_EMITS;
    if (type === "props")
      hasDefineProps = true;
    else
      hasDefineEmits = true;
    if (node.arguments[0]) {
      throw new SyntaxError("error: 2");
    }
    const typeDeclRaw = (_b = (_a = node.typeParameters) == null ? void 0 : _a.params) == null ? void 0 : _b[0];
    if (!typeDeclRaw) {
      throw new SyntaxError("Error: 3");
    }
    const typeDecl = resolveQualifiedType(
      typeDeclRaw,
      (node2) => node2.type === "TSTypeLiteral"
    );
    if (!typeDecl) {
      throw new Error(
        `type argument passed to ${fnName}() must be a literal type, or a reference to an interface or literal type.`
      );
    }
    if (type === "props")
      propsTypeDecl = typeDecl;
    else
      emitsTypeDecl = typeDecl;
    if (declId) {
      if (type === "props" && declId.type === "ObjectPattern") {
        propsDestructureDecl = declId;
      } else if (type === "emits" && declId.type === "Identifier") {
        emitsIdentifier = declId.name;
      }
    }
    return true;
  }
  function processDefineModel(node, declId, kind) {
    var _a;
    if (!isCallOf(node, DEFINE_MODEL)) {
      return false;
    }
    if (hasDefineModel) {
      throw new SyntaxError(`duplicate ${DEFINE_MODEL}() call`);
    }
    hasDefineModel = true;
    const propsTypeDeclRaw = (_a = node.typeParameters) == null ? void 0 : _a.params[0];
    if (!propsTypeDeclRaw) {
      throw new SyntaxError("Error: 1");
    }
    modelTypeDecl = resolveQualifiedType(
      propsTypeDeclRaw,
      (node2) => node2.type === "TSTypeLiteral"
    );
    if (!modelTypeDecl) {
      throw new Error(
        `type argument passed to ${DEFINE_MODEL}() must be a literal type, or a reference to an interface or literal type.`
      );
    }
    if (declId) {
      const ids = extractIdentifiers(declId);
      ids.forEach((id) => modelIdentifiers.add(id));
      if (declId.type === "ObjectPattern") {
        modelDestructureDecl = declId;
        for (const property of declId.properties) {
          if (property.type === "RestElement") {
            throw new SyntaxError("not supported");
          }
        }
      } else {
        modelIdentifier = scriptSetup.loc.source.slice(
          declId.start,
          declId.end
        );
      }
    }
    if (kind)
      modelDeclKind = kind;
    return true;
  }
  function resolveQualifiedType(node, qualifier) {
    if (qualifier(node)) {
      return node;
    }
    if (node.type === "TSTypeReference" && node.typeName.type === "Identifier") {
      const refName = node.typeName.name;
      const isQualifiedType = (node2) => {
        if (node2.type === "TSInterfaceDeclaration" && node2.id.name === refName) {
          return node2.body;
        } else if (node2.type === "TSTypeAliasDeclaration" && node2.id.name === refName && qualifier(node2.typeAnnotation)) {
          return node2.typeAnnotation;
        } else if (node2.type === "ExportNamedDeclaration" && node2.declaration) {
          return isQualifiedType(node2.declaration);
        }
      };
      const body = sfc.scriptSetup.scriptSetupAst;
      for (const node2 of body) {
        const qualified = isQualifiedType(node2);
        if (qualified) {
          return qualified;
        }
      }
    }
  }
  function extractRuntimeProps(node) {
    const members = node.type === "TSTypeLiteral" ? node.members : node.body;
    const map2 = {};
    for (const m of members) {
      if ((m.type === "TSPropertySignature" || m.type === "TSMethodSignature") && m.key.type === "Identifier") {
        const value = scriptSetup.loc.source.slice(
          m.typeAnnotation.start,
          m.typeAnnotation.end
        );
        map2[m.key.name] = value;
      }
    }
    return map2;
  }
  function getEventKey(key) {
    if (key === "value" && version === 2)
      return "input";
    return `update:${key}`;
  }
  function processEmitValue() {
    if (!emitsIdentifier)
      throw new Error("Error: 4");
    const program = {
      type: "Program",
      body: scriptSetup.scriptSetupAst,
      directives: [],
      sourceType: "module",
      sourceFile: ""
    };
    let hasTransfromed = false;
    walkAST(program, {
      leave(node) {
        if (node.type !== "AssignmentExpression")
          return;
        if (node.left.type !== "Identifier")
          return;
        const idDecl = this.scope[node.left.name];
        if (!modelIdentifiers.has(idDecl))
          return;
        hasTransfromed = true;
        s.overwrite(
          startOffset + node.start,
          startOffset + node.end,
          `__emitHelper(${emitsIdentifier}, '${getEventKey(
            idDecl.name
          )}', ${s.slice(
            startOffset + node.right.start,
            startOffset + node.right.end
          )})`
        );
      }
    });
    if (hasTransfromed) {
      s.prependLeft(
        startOffset,
        "import { emitHelper as __emitHelper } from 'unplugin-vue-macros/helper'\n"
      );
    }
  }
  if (!code.includes(DEFINE_MODEL))
    return;
  const sfc = parseSFC(code, filename);
  if (!sfc.scriptSetup)
    return;
  if (!sfc.scriptSetup.scriptSetupAst) {
    sfc.scriptSetup = compileScript(sfc, {
      id: filename
    });
  }
  const { scriptSetup } = sfc;
  const startOffset = scriptSetup.loc.start.offset;
  const s = new MagicString(code);
  for (const node of scriptSetup.scriptSetupAst) {
    if (node.type === "ExpressionStatement") {
      processDefinePropsOrEmits(node.expression);
      if (processDefineModel(node.expression)) {
        s.remove(node.start + startOffset, node.end + startOffset);
      }
    }
    if (node.type === "VariableDeclaration" && !node.declare) {
      const total = node.declarations.length;
      let left = total;
      for (let i = 0; i < total; i++) {
        const decl = node.declarations[i];
        if (decl.init) {
          processDefinePropsOrEmits(decl.init, decl.id);
          if (processDefineModel(decl.init, decl.id, node.kind)) {
            if (left === 1) {
              s.remove(node.start + startOffset, node.end + startOffset);
            } else {
              let start = decl.start + startOffset;
              let end = decl.end + startOffset;
              if (i < total - 1) {
                end = node.declarations[i + 1].start + startOffset;
              } else {
                start = node.declarations[i - 1].end + startOffset;
              }
              s.remove(start, end);
              left--;
            }
          }
        }
      }
    }
  }
  if (!modelTypeDecl)
    return;
  if (modelTypeDecl.type !== "TSTypeLiteral") {
    throw new SyntaxError(
      `type argument passed to ${DEFINE_MODEL}() must be a literal type, or a reference to an interface or literal type.`
    );
  }
  const map = extractRuntimeProps(modelTypeDecl);
  const propsText = Object.entries(map).map(([key, value]) => `${key}${value}`).join("\n");
  const emitsText = Object.entries(map).map(([key, value]) => `(evt: '${getEventKey(key)}', value${value}): void`).join("\n");
  if (hasDefineProps) {
    s.appendLeft(startOffset + propsTypeDecl.start + 1, `${propsText}
`);
    if (propsDestructureDecl && modelDestructureDecl)
      for (const property of modelDestructureDecl.properties) {
        const text = code.slice(
          startOffset + property.start,
          startOffset + property.end
        );
        s.appendLeft(startOffset + propsDestructureDecl.start + 1, `${text}, `);
      }
  } else {
    let text = "";
    const kind = modelDeclKind || "let";
    if (modelIdentifier) {
      text = modelIdentifier;
    } else if (modelDestructureDecl) {
      text = code.slice(
        startOffset + modelDestructureDecl.start,
        startOffset + modelDestructureDecl.end
      );
    }
    s.appendLeft(
      startOffset,
      `${text ? `${kind} ${text} = ` : ""}defineProps<{
  ${propsText}
}>();
`
    );
  }
  if (hasDefineEmits) {
    s.appendLeft(startOffset + emitsTypeDecl.start + 1, `${emitsText}
`);
  } else {
    emitsIdentifier = `_${DEFINE_MODEL}_emit`;
    s.appendLeft(
      startOffset,
      `const ${emitsIdentifier} = defineEmits<{
  ${emitsText}
}>();
`
    );
  }
  if (hasDefineModel) {
    processEmitValue();
  }
  return s;
};

export {
  transformDefineModel
};
